/*
** Module   : ChuiWebSimulator.java
** Abstract : A WebSocket Chui Terminal simulator.
**
** Copyright (c) 2013-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**         Golden Code Development Corporation
**                 CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 MAG 20131113 Created initial version.
** 002 MAG 20140114 Implements WebSockets protocol and communication.
** 003 MAG 20140204 Send text attributes via WebSockets. Add a Watchdog Timer 
**                  guardian and quit command.                   
** 004 MAG 20140211 Implements COLOR.
** 005 CA  20140206 Web socket timeout can't be read from the jetty threads (as these don't have a
**                  P2J context, so it must be passed at construction.
** 006 MAG 20140217 Refactoring web client redirection flaw.
** 007 MAG 20140219 Improve performances.
** 008 MAG 20140402 Implements interactive process.
*/

package com.goldencode.p2j.ui.client.chui.driver.web;

import java.io.*;
import java.nio.*;
import java.util.*;
import java.util.concurrent.*;

import org.eclipse.jetty.websocket.api.*;
import org.eclipse.jetty.websocket.api.annotations.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.chui.driver.swing.*;

/**
 * Chui WebSocket terminal simulator.
 * Implements the WebSockets protocol to communicate with the JS terminal simulator.
 * <p>
 * IMPORTANT: Do not use logging anywhere inside this class. When an authentication plugin
 * is running during session establishment with the P2J server using logging no messages are 
 * written into the log and the application is blocked.
 * <p>   
 * The payloads are in 2 formats binary and JSON text.
 * <p>
 * Input messages:
 * 
 * Keyboard input.
 * 
 * offset  value content
 * ------------------------------------
 * 0       0x01  message id
 * 1       uu    keystroke int value
 * 2       ll    keystroke = uu << 8 | ll 
 * <p>
 * Keyboard pair input.
 * 
 * offset  value content
 * ------------------------------------
 * 0       0x02  message id
 * 1       uu    keyCode int value
 * 2       ll    keyCode = uu << 8 | ll 
 * 3       uu    charCode int value
 * 4       ll    charCode = uu << 8 | ll 
 * <p>
 * Output messages:
 * 
 * Clear screen.
 * 
 * offset  value content
 * ------------------------------------
 * 0       0x80  message id
 * <p>
 * Draw text on screen.
 * 
 * offset  value content
 * ------------------------------------
 * 0       0x81  message id
 * 1..n    data  data packets
 * 
 * structure of data packets:

 * offset  value  content
 * ------------------------------------
 * 0       size   packet size high byte
 * 1       size   packet size low byte
 * 2       width  rectangle with
 * 3       height rectangle height
 * 4       row    row number
 * 5       col    column number
 * 6       attr   text attributes
 * 7       color  FG/BG color pairs
 * 8..n    text   text characters, each character on 2 bytes (UTF-16)
 * <p>
 * Set cursor position.
 * 
 * offset  value content
 * ------------------------------------
 * 0       0x82  message id
 * 1       row   row
 * 2       col   column
 * <p>
 * Set cursor status.
 * 
 * offset  value content
 * ------------------------------------
 * 0       0x83  message id
 * 1       stat  status 1 = visible 
 *                      0 = hidden 
 * <p>
 * Play a beep message.
 * 
 * offset  value content
 * ------------------------------------
 * 0       0x84  message id
 * <p>
 * Quit message.
 * 
 * offset  value content
 * ------------------------------------
 * 0       0x85  message id
 * <p>
 * Palette message is in JSON format as follow.
 * 
 *    {"c":0, "p":[color1, color2, ... color3]}
 * 
 * Example:
 *    {"c":0, "p":["#000000", "#FF0000", "#00FF00", "#0000FF", "#FFFFFF"]}
 * 
 * offset  value content
 * ------------------------------------
 * 0       0x86  message id
 * 1       mode  mode 1 = VT100 
 *                    0 = P2J 
 * <p>
 * Switch mode.
 * 
 */
@SuppressWarnings("serial")
@WebSocket(maxIdleTime = 900000, maxTextMessageSize = 4096, maxBinaryMessageSize = 32768)
public class ChuiWebSimulator
extends ChuiSimulator
{  
   /** keyboard stroke message */
   private static final byte MSG_KEY           = (byte) 0x01;

   /** keyboard stroke message */
   private static final byte MSG_KEY_VT100     = (byte) 0x02;

   /** clear screen message */
   private static final byte MSG_CLEAR         = (byte) 0x80;

   /** draw text on screen */
   private static final byte MSG_DRAW          = (byte) 0x81;
   
   /** set cursor position */
   private static final byte MSG_CURSOR_POS    = (byte) 0x82;
   
   /** set cursor status */
   private static final byte MSG_CURSOR_STATUS = (byte) 0x83;
   
   /** message beep */
   private static final byte MSG_BEEP          = (byte) 0x84;
   
   /** quit message */
   private static final byte MSG_QUIT          = (byte) 0x85;

   /** switch to VT100 mode */
   private static final byte MSG_VT100         = (byte) 0x86;
   
   /** Color palette */
   private static final String COLOR_PALETTE = "{\"c\":0, \"p\":%s}"; 
   
   /** Keyboard input queue */
   private Deque<Integer> keyBuffer = new ConcurrentLinkedDeque<>();
   
   /** Keep cursor status */
   private boolean cursorVisible = true;
   
   /** Web socket timeout. */
   private final long timeout;
   
   /** Watchdog time out value */
   private final long watchdogTimeout;

   /** Watchdog instance */
   private WatchdogTimer watchdog;
   
   /** Color palette JSON message */
   private String paletteMessage;
   
   /** Number of rows */
   private int rows;
   
   /** Number of columns */
   private int columns;

   /** Screen contents. */
   private Cell[][] cache = null;
     
   /** Push messages worker thread */
   private PushMessagesWorker pushWorker;
   
   /** Number of simultaneous connections */
   private volatile int connections = 0;
   
   /** Store terminal mode */
   private boolean isVT100 = false;
         
   /**
    * Create a new chui web simulator.
    * 
    * @param    timeout
    *           Web socket timeout.
    * @param    watchdogTimeout
    *           Watchdog thread timeout.
    */
   public ChuiWebSimulator(long timeout, long watchdogTimeout)
   {
      // build palette message
      this.paletteMessage = buildPaletteMessage();
      this.timeout = timeout;
      this.watchdogTimeout = watchdogTimeout;
      
      // Start watchdog timer
      startWatchdog(watchdogTimeout);
   }
   
   /**
    * This method is called when the remote peer open a web socket connection.
    *
    * @param   session
    *          An active link of communications with a Remote WebSocket Endpoint.
    *
    */
   @OnWebSocketConnect
   public void onConnect(Session session)
   {
      synchronized (keyBuffer)
      {
         if (++connections == 1)
         {
            // Stop watchdog timer
            stopWatchdog();
            // Set session timeout.
            if (timeout > 0)
            {
               session.setIdleTimeout(timeout);
            }
            // Start push worker thread
            startPushWorker(session);
            // synchronize tty
            initRemoteTerminal();
         }
         else
         {
            // multiple connections not allowed 
            session.close();
         }
      }
   }

   /**
    * This method is called when a text message is received.
    *
    * @param   message
    *          The text message.
    */
   @OnWebSocketMessage
   public void onMessage(String message) 
   {
      // Not used.
   }

   /**
    * This method is called when a binary message is received.
    *
    * @param   message
    *          Message content as an byte array.
    * @param   offset
    *          Message offset..
    * @param   length
    *          Message size.
    */
   @OnWebSocketMessage
   public void onMessage(byte[] message, int offset, int length)
   {
      if (length == 3 && message[offset] == ChuiWebSimulator.MSG_KEY)
      {
         // keystroke
         int keystroke = (message[offset + 1] << 8) & 0xFF00;
         keystroke |= (message[offset + 2] & 0x00FF); 
         keyBuffer.offerLast(keystroke);
      }
      else if (length == 5 && message[offset] == ChuiWebSimulator.MSG_KEY_VT100)
      {
         // keystroke
         int keyCode = (message[offset + 1] << 8) & 0xFF00;
         keyCode |= (message[offset + 2] & 0x00FF); 
         int charCode = (message[offset + 3] << 8) & 0xFF00;
         charCode |= (message[offset + 4] & 0x00FF); 
         this.injectKey(keyCode, (char) charCode);
      }
   }

   /**
    * This method is called when the connection is closed.
    *
    * @param   statusCode
    *          The status integer code.
    * @param   reason
    *          A string representing the reason.
    */
   @OnWebSocketClose
   public void onClose(int statusCode, String reason)
   {
      synchronized (keyBuffer)
      {
         if (--connections == 0)
         {
            // Stop push worker thread
            stopPushWorker();
            // Start watchdog timer
            startWatchdog(watchdogTimeout);            
         }
      }
   }
   
   /**
    * This method is called on communication errors.
    *
    * @param   session
    *          An active link of communications with a Remote WebSocket Endpoint.
    * @param   error
    *          A <code>Throwable</code> instance representing the error condition.
    */
   @OnWebSocketError
   public void onError(Session session, Throwable error)
   {
      synchronized (keyBuffer)
      {
         session.close();      
      }
   }

   /**
    * Start watchdog timer.
    * 
    * @param   watchdogTimeout
    *          Watchdog timeout value.
    */
   private void startWatchdog(long watchdogTimeout)
   {
      if (watchdog == null)
      {
         watchdog = new WatchdogTimer(watchdogTimeout);
      }
   }

   /**
    * Stop watchdog timer.
    */
   private void stopWatchdog()
   {
      if (watchdog != null)
      {
         watchdog.kill();
         watchdog = null;
      }
   }

   /**
    * Start pushWorker thread.
    * 
    * @param   session
    *          An active link of communications with a Remote WebSocket end point.
    */
   private void startPushWorker(Session session)
   {
      if (pushWorker == null)
      {
         pushWorker = new PushMessagesWorker(session);
      }
   }

   /**
    * Stop pushWorker thread.
    */
   private void stopPushWorker()
   {
      if (pushWorker != null)
      {
         pushWorker.stopPushing();
         pushWorker = null;
      }
   }
   
   /**
    * Send a text message.
    * 
    * @param   message
    *          Text message.
    */   
   public void sendTextMessage(String message) 
   {
      synchronized (keyBuffer)
      {
         if (pushWorker != null)
         {
            pushWorker.pushMessage(message);
         }
      }
   }

   /**
    * Send a binary message.
    * 
    * @param   message
    *          Binary message.
    */   
   public void sendBinaryMessage(byte[] message) 
   {
      synchronized (keyBuffer)
      {
         if (pushWorker != null)
         {
            pushWorker.pushMessage(ByteBuffer.wrap(message));
         }
      }
   }

   /**
    * Whenever the main page is loaded on browser synchronize virtual terminal state.
    */
   private void initRemoteTerminal()
   {
      // clear cache
      cache = new Cell[rows][columns];
      colorPalette();
      // synchronize
      sendSwitchMode();
      triggerRepaint(true);
      setCursorStatus(cursorVisible);
      if (cursorVisible)
      {
         setCursor(getCursorX(), getCursorY());
      }
   }
   
   /**
    * Read the next key from the FIFO key input buffer.  
    *
    * @return   The next key in the buffer or -1 if buffer is empty.
    */
   @Override
   public int readKey()
   {
      int result = -1;
      if (keyBuffer != null && !keyBuffer.isEmpty())
      {
         Integer key = keyBuffer.pollFirst();
         if (key != null)
         {
            result = key.intValue();
         }
      }
      return result;
   }
   
   /**
    * Initialize the instance using the given configuration.
    * Allocate screen cache. 
    *
    * @param   options
    *          The configuration for the terminal simulator.
    */   
   @Override
   public void configure(TerminalOptions options)
   {
      super.configure(options);
      
      synchronized (keyBuffer)
      {
         columns = options.columns;
         rows = options.rows;
      
         if (cache == null)
         {
            cache = new Cell[rows][columns];
         }
      }
   }

   /**
    * Overwrite the entire 2-dimensional output buffer with the space character
    * to be rendered with the <code>Color.NORMAL</code> color. Sets the cursor
    * Y and X positions to 0. Send INIT message down to JS client.
    */   
   @Override
   public void clear()
   {
      super.clear();
      ByteArrayOutputStream message = new ByteArrayOutputStream();
      message.write(ChuiWebSimulator.MSG_CLEAR);
      sendBinaryMessage(message.toByteArray());
   }

   /**
    * The beep() should be implemented by sending a command on the remote end point.
    * The command is part of the communication protocol.
    *
    */
   public void beep()
   {
      ByteArrayOutputStream message = new ByteArrayOutputStream();
      message.write(ChuiWebSimulator.MSG_BEEP);
      sendBinaryMessage(message.toByteArray());         
   }
   
   /**
    *  Set the cursor visibility. Send SET_CURSOR_ON or SET_CURSOR_OFF message down to JS client.
    *
    * @param    on
    *           <code>true</code> to make the cursor visible (subsequent repaint
    *           operations will then cause the cursor to be drawn).
    *
    * @return   The previous cursor visibility state.
    */   
   @Override
   public boolean setCursorStatus(boolean on) 
   throws IllegalStateException
   {
      cursorVisible = on;
      ByteArrayOutputStream message = new ByteArrayOutputStream();
      message.write(ChuiWebSimulator.MSG_CURSOR_STATUS);
      message.write(on ? 1 : 0);
      sendBinaryMessage(message.toByteArray());         
      // call super implementation
      return super.setCursorStatus(on);
   }

   /**
    * Position the cursor. Send SET_CURSOR_POS message down to JS client.
    *
    * @param    x
    *           New X cursor position.
    * @param    y
    *           New Y cursor position.
    *
    * @return  <code>true</code> if cursor positioning succeeded.
    */
   @Override
   public boolean setCursor(int x, int y)
   {
      ByteArrayOutputStream message = new ByteArrayOutputStream();
      message.write(ChuiWebSimulator.MSG_CURSOR_POS);
      message.write(y);
      message.write(x);
      // send message
      sendBinaryMessage(message.toByteArray());         
      // call super implementation
      return super.setCursor(x, y);
   }
   
   /**
    * Put cell attributes and character into the message buffer.
    * 
    * @param   message
    *          Message buffer.
    * @param   cellColor
    *          Cell color object.
    */
   private void putCellAttributes(ByteArrayOutputStream message, Color cellColor)
   {
      // attributes
      if (cellColor == null)
      {
         // attributes
         message.write(0);
         // color
         message.write(0);                        
      }
      else
      {
         // attributes
         message.write(cellColor.attribute());
         // color
         int color = cellColor.color();
         if (color == 0)
         {
            message.write(0);               
         }
         else
         {
            int bg = color & 0x07;
            int fg = (color >>> 12) & 0x70;
            message.write(fg | bg);
         }               
      }
   }
   
   /**
    * Put cell character into the message buffer.
    * 
    * @param   message
    *          Message buffer.
    * @param   cell
    *          Cell object.
    */
   private void putCellCharacter(ByteArrayOutputStream message, Cell cell)
   {
      // character
      char c = cell.character();
      message.write((c >>> 8) & 0xFF);
      message.write(c & 0xFF);
   }
   
   /**
    * When data is prepared to send down to browser JS code to be draw the data should be 
    * optimized for drawing on canvas.
    * <p>
    * The video screen memory is composed form cells stored in a 2 dimensional matrix,
    * The size of the matrix is rows x columns where rows are the number of rows in the screen
    * and columns are the number of columns. For example 24 x 80 for a screen having 24 rows 
    * and 80 columns. Each cell is a structure which holds cell attributes, color and character.
    * The status of video screen memory is cached and preserved the last screen status.
    * Whenever a new screen should be send down the new screen is compared with the cached screen
    * and only the changed cells are send down. Also the cached status is updated to reflect the 
    * latest changes. The cell is considered changed when at least one of the cell properties 
    * attribute, color or character had changed.
    * <p>
    * Drawing texts on HTML5 canvas using JS code is done on two steps:
    * <ul> 
    *  <li> First the background rectangle containing the text is filled with the background color. 
    *       The largest the filled rectangle is the drawing is optimal.
    *       
    *  <li> The second step is to fill the rectangle with the containing text using the foreground 
    *       color. On HTML5 canvas this can be done only on a line basis by filling with text 
    *       rectangles having the height equals to one character. An optimized draw on this case 
    *       require a text length as much as possible long.
    * </ul>
    * <p>
    * When a network message is constructed this is done on two steps by keeping in mind the 
    * performances constraints described above.
    * <ul>
    *  <li> The new screen that have to be send is scanned line by line column by column to detect 
    *  changed cells. Adjacent changed cells on a line having the same color and attributes are 
    *  stored together on a Chunk structure. The chunks are stored on a linked map structure having 
    *  the cell Color (attribute and color)  structure as a key and a linked list of chunks as value. 
    *  Both the linked map and the linked list preserve the inserting order.
    *  
    *  <li> On the next step the message is build and contains all chunks from a screen detected on 
    *  the first step.  For each entry on the map the list of chunks is used to build rectangle 
    *  structures. The chunks are extracted from the list one by one. If multiple chunks could be 
    *  combined on larger rectangles the chunks are combined to form a largest rectangle. The width 
    *  of a rectangle is the number of characters from a chunk and the height are the number of chunks 
    *  combined. For example a rectangle  of 25x5 holds 5 lines of 25 text characters each. 
    *  All characters have the same color and text attributes. The condition for two chunks to form a 
    *  rectangle is to have the same length, to start at the same position within the line and are from
    *  adjacent lines. Chunks that cannot be combined as large rectangles are stored inside the message
    *  as simple rectangles, rectangles having height equals to one.
    * </ul>
    * 
    * @param    full
    *           Flag to force a the entire screen to repaint.
    */
   @Override
   public void triggerRepaint(boolean full)
   {
      synchronized (keyBuffer)
      {
         /* chunks map */
         Map<Color, List<Chunk>> chunks = new LinkedHashMap<>();
         
         // chunk
         ByteArrayOutputStream chunk = null;
         
         // get screen data
         Cell[][] data = getScreenData();
               
         for (int row = 0; row < rows; row++)
         {             
            // attributes of previous cell
            Color previous = null;

            for (int col = 0; col < columns; col++)
            {
               Cell cell = data[row][col];
               
               // check if cell has changes
               if (cell != null && !cell.equals(cache[row][col]))
               {
                  Color color = cell.color(); 

                  // check for color and attributes changes
                  if (color != null && color.equals(previous))
                  {
                     // same color and attributes.
                     // add cell to the current chunk.
                     putCellCharacter(chunk, cell);
                  }
                  else
                  {
                     // attributes changed.
                     // add current chunk if exists and start a new one.
                     addChunk(chunks, chunk, previous);
                     // new chunk
                     previous = color;
                     chunk = new ByteArrayOutputStream();
                     chunk.write(row); // row number                     
                     chunk.write(col); // column number
                     putCellCharacter(chunk, cell);
                  }
               }
               else
               {
                  // the cell is already in cache do not send. 
                  // add current chunk and reinitialize. 
                  addChunk(chunks, chunk, previous);
                  previous = null;
               }
               
               cache[row][col] = cell;               
            }
            // start a new line
            // check if a chunk exists and add to the list of chunks.
            addChunk(chunks, chunk, previous);
         }   
         
         // build the message from chunks map.
         if (!chunks.isEmpty())
         {
            buildMessage(chunks);
         }
      }
   }
   
   /**
    * Add a chunk to the chunks map.
    * 
    * @param   chunks
    *          Chunks map.
    * @param   chunk
    *          Chunk to add.
    * @param   color
    *          Color attribute used as a key
    */
   private void addChunk(Map<Color, List<Chunk>> chunks, 
                         ByteArrayOutputStream chunk,
                         Color color)
   {
      if (color != null && chunk != null && chunk.size() > 0)
      {
         List<Chunk> values = chunks.get(color);
         if (values == null)
         {
            values = new LinkedList<Chunk>();
            chunks.put(color, values);
         }
         values.add(new Chunk(chunk.toByteArray()));
      }
   }
   
   /**
    * Build the message from chunks map.
    *  
    * @param   chunks
    *          Chunks map.
    */
   private void buildMessage(Map<Color, List<Chunk>> chunks)
   {
      // message
      ByteArrayOutputStream message = new ByteArrayOutputStream();
      message.write(ChuiWebSimulator.MSG_DRAW);
      
      for (Map.Entry<Color, List<Chunk>> entr : chunks.entrySet())
      {
         Color color = entr.getKey();
         // list of chunks having the same color
         LinkedList<Chunk> list = (LinkedList<Chunk>) entr.getValue();
         // build a message as rectangles
         while (!list.isEmpty())
         {
            // rectangle height
            int height = 1;
            // poll first chunk
            Chunk chunk = list.pollFirst();
            // try to build a rectangle
            ByteArrayOutputStream rect = new ByteArrayOutputStream();
            rect.write(0); // packet size
            rect.write(0);
            rect.write(chunk.getTextSize() / 2);  // width
            rect.write(1);  // height
            rect.write(chunk.getRow());  // row
            rect.write(chunk.getColumn());  // col
            // add attributes and color
            putCellAttributes(rect, color);
            // write text
            rect.write(chunk.getChunk(), 
                       Chunk.HEADER_SIZE, 
                       chunk.getTextSize());
            // check on the next line
            for (int i = chunk.getRow() + 1; i < rows; i++)
            {
               boolean rectangleDetected = false;
               
               Chunk next = new Chunk(i, chunk.getColumn(), chunk.getSize());
               
               // search a candidate to add to rectangle
               for (Chunk candidate : list)
               {
                  // detect rectangle 
                  rectangleDetected = candidate.equals(next); 

                  // if found rectangle
                  if (rectangleDetected)
                  {
                     height++;
                     // consolidate rectangle
                     rect.write(candidate.getChunk(), 
                                Chunk.HEADER_SIZE, 
                                candidate.getTextSize());
                     list.remove(candidate);
                     break;
                  }               
               }               
               
               // end of rectangle detection. 
               if (!rectangleDetected)
               {
                  break;
               }
            }
            // update packet size, rectangle height and add to message.
            byte[] rectBytes = rect.toByteArray();
            int packetSize = rectBytes.length;
            rectBytes[0] = (byte)((packetSize >>> 8) & 0xFF);
            rectBytes[1] = (byte) (packetSize & 0xFF);
            rectBytes[3] = (byte) height;
            message.write(rectBytes, 0, rectBytes.length);
         }
      }
      
      // put the message on queue
      sendBinaryMessage(message.toByteArray());
   }
   
   /**
    * Remote application is about to end. Notify the JS client about that.
    * The client should be redirected to referrer address.
    */
   public void quit()
   {
      ByteArrayOutputStream message = new ByteArrayOutputStream();
      message.write(ChuiWebSimulator.MSG_QUIT);
      sendBinaryMessage(message.toByteArray());         
   }

   /**
    * Build color palette message. This message is the same
    * over the user session so it's build and store on constructor.
    * The message is send whenever a WebSocket connection is open.
    *
    * @return  A string representing the JSON format of color palette.
    */
   private String buildPaletteMessage()
   {
      // build palette
      StringBuilder palette = new StringBuilder("[");      
      for (ColorPalette colorPalette : ColorPalette.values())
      {
         if (palette.length() != 1)
         {
            palette.append(",");
         }
         palette.append("\"").append(colorPalette.asWebColor()).append("\"");
      }      
      palette.append("]");      
      // JSON format
      return String.format(COLOR_PALETTE, palette.toString());
   }
   
   /**
    * Send color palette down to JS client.
    * This is done whenever a WebSocket connection is open.
    */
   private void colorPalette()
   {      
      // send palette
      sendTextMessage(paletteMessage);
   }
   
   /**
    * Switch mode message.
    */
   private void sendSwitchMode()
   {
      ByteArrayOutputStream message = new ByteArrayOutputStream();
      message.write(ChuiWebSimulator.MSG_VT100);
      message.write(this.isVT100 ? 1 : 0);
      sendBinaryMessage(message.toByteArray());               
   }
   
   /**
    * Switch between P2J client and interactive client.
    * 
    * @param   toVT100
    *          If <code>true</code> show interactive process screen otherwise P2J client screen.
    */
   @Override
   public void swapKeyboards(boolean toVT100)
   {
      this.isVT100 = toVT100;
      // Send message
      this.sendSwitchMode();
  }   
}