/*
** Module   : VT100Emulator.java
** Abstract : DEC VT-100 emulator.
**
** Copyright (c) 2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ----------------Description-----------------
** 001 MAG 20140318 First version.
*/

package com.goldencode.terminal;

import java.io.*;
import java.util.*;

/**
 * This is an implementation of DEC VT-100 specification.
 * 
 * @author mag
 *
 */
public class VT100Emulator 
extends MasterPty
implements Runnable
{
   /** Terminal type */
   private static final String VT_100 = "vt100";
     
   /** Escape code */
   private static final byte ESC = (byte) 0x1b;

   /** DA response */
   private static final byte[] DA_RESPONSE = {ESC, '[', '?', '1', ';', '0', 'c'};
   
   /** Identify terminal response */
   private static final byte[] IDENT_RESPONSE = {ESC, '/', 'Z'};
   
   /** Terminal mode */
   private TerminalMode mode = TerminalMode.ANSI;
   
   /** Cell attributes and colors */
   private int attribute = 0, color = 0;
   
   /** attribute storage */
   private AttributeStorage attrStorage = null;
   
   /** Tab stops */
   private Set<Integer> tabStops = new TreeSet<Integer>();
            
   /** Main loop is running */
   private volatile boolean running = false;

   /** Request terminate */
   private volatile boolean terminate = false;
   
   /** VT-100 keyboard */
   protected VT100Keyboard keyboard;

   /** Master PTY interface */
   protected VT100Renderer pty;
   
   /**
    * Constructor. Used when switching between P2J and emulator.
    * 
    * @param   pty
    *          A VT-100 renderer. 
    * @param   keyboard
    *          A VT-100 keyboard handler. 
    * @param   cmdLine
    *          Command line arguments. 
    * @param   overlay
    *          Overlapped mode. 
    */
   public VT100Emulator(VT100Renderer pty, 
                        VT100Keyboard keyboard, 
                        String[] cmdLine,
                        int overlay)
   {
      super(VT100Emulator.VT_100, cmdLine, overlay);
      this.pty = pty;
      this.keyboard = keyboard;
      this.keyboard.setMasterPty(this);
   }
   
   /**
    * Main loop. The thread should exit when no bytes are available
    * to read form master pty and a request to terminate is on place.
    */
   public void doWork()
   {
      while (running)
      {
         byte b = this.nextByte();
         switch (b)
         {
         case 0x00:
         case 0x7f:
            // ignored
            break;
         case (byte) 0xff:
            // no bytes to read
            running = !terminate;
            break;
         case 0x07:
            // bell
            pty.beep();
            break;
         case 0x08:
            // BACKSPACE cursor left
            pty.moveCursorPos(0, -1);
            break;
         case 0x09:
            // tab
            this.goNextTabstop();
            break;
         case 0x0a:
            // LF cursor down
            pty.nextLine();
            break;
         case 0x0c:
            // Formfeed (also: New page)
            pty.eraseInDisplay(2);
            pty.setCursorPos(0, 0);
            break;
         case 0x0d:     
            // CR Carriage return
            pty.setCursorCol(0);
            break;
         case 0x0e:
            // smacs ^N // start alternate character set (P)
            break;
         case 0x0f:
            // rmacs ^O // end alternate character set (P)
            break;
         case ESC:
            // ESCAPE sequence
            doEscSequence();
            break;
         default:
            //FIX UTF
            putUtf8Stream((int) b & 0xff);
            break;
         }
      }
   }

   /**
    * Process ESC sequences
    */
   private void doEscSequence()
   {
      if (mode == TerminalMode.ANSI)
      {
         doAnsi();
      } 
      else
      {
         doVT52();
      }      
   }
   
   /**
    * Process ANSI mode
    */
   private void doAnsi()
   {
      byte b = this.nextByte();
      switch (b)
      {
      case '<':   
         // SETANSI Enter/exit ANSI mode (VT52)
         mode = TerminalMode.VT52;
         break;
      case '[':
         processCSI();
         break;
      case 'H':
         // HTS - Set horizontal tab stop
         tabStops.add(pty.getCursorCol());
         break;
      case '=':
         // DECKPAM Set alternate keypad mode
         break;
      case '>':
         // DECKPNM Set numeric keypad mode
         break;
      case 'D':
         // IND Move/scroll window up one line
         pty.index();
         break;
      case 'M':
         // RI Move/scroll window down one line
         pty.reverseIndex();
         break;
      case 'E':
         // NEL Move to next line
         pty.nextLine();
         break;
      case '7':
         // DECSC Save cursor position and attributes
         attrStorage = new AttributeStorage(pty.getCursorRow() + 1, 
                                            pty.getCursorCol() + 1, 
                                            attribute, 
                                            color);
         break;
      case '8':
         // DECSC Restore cursor position and attributes
         if (attrStorage != null)
         {
            attribute = attrStorage.attribute;
            color = attrStorage.color;
            pty.setCursorPos(attrStorage.row, attrStorage.col);
         }
         break;
      case 'Z':
         // DA -- Device Attributes ESC [?1;0c         
         put(DA_RESPONSE);
         break;
      case ')':
      case '(':
         // SCS -- Select Go and G1 Character Set 
         b = this.nextByte();
         break;
      case 'N':
      case 'O':
         // SS2 -- Select G2 and G3 Character Set
         break;
      default:
         break;
      }
   }

   /**
    * process ESC [ Pn;...Pn x
    */
   private void processCSI()
   {
      byte b;
      int argc = 0;
      int[] argn = new int[32];
      
      // decode numeric parameters
      while (true)
      {
         b = this.nextByte();
         if (b == ';')
         {
            argc++;
            argn[argc] = 0;
         }
         else if (isDigit(b))
         {
            argn[argc] = argn[argc] * 10 + (b - 0x30);
         }
         else
         {
            // stop
            break;            
         }
      }
      
      // for cursor functions
      int arg0 = argn[0] == 0 ? 1 : argn[0];

      switch (b)
      {
      // CUU -- Cursor Up
      case 'A':
         pty.moveCursorPos(-arg0, 0);
         break;
      // CUD -- Cursor Down
      case 'B':
         pty.moveCursorPos(arg0, 0);
         break;
      // CUF -- Cursor Forward
      case 'C':
         pty.moveCursorPos(0, arg0);
         break;
      // CUB -- Cursor Backward
      case 'D':
         pty.moveCursorPos(0, -arg0);
         break;
      // HVP -- Horizontal and Vertical Position
      case 'f':
      // CUP -- Cursor Position
      case 'H':
         pty.setCursorPos(argn[0], argn[1]);
         break;
      // ED -- Erase In Display (default 0)
      case 'J':
         switch (argn[0]) 
         {
         case 0:
            // Erase from the active position to the end of the screen, inclusive (default)
            pty.eraseInDisplay(0);
            break;
         case 1:
            // Erase from start of the screen to the active position, inclusive
            pty.eraseInDisplay(1);
            break;
         case 2:
            // Erase all of the display -- all lines are erased, changed to single-width, and the cursor does not move.
            pty.eraseInDisplay(2);
            break;
         }
         break;
      // EL -- Erase In Line (default 0)
      case 'K':
         switch (argn[0]) 
         {
         case 0:
            // Erase from the active position to the end of the line, inclusive (default)
            pty.eraseInLine(0);
            break;
         case 1:
            // Erase from the start of the screen to the active position, inclusive
            pty.eraseInLine(1);
            break;
         case 2:
            // Erase all of the line, inclusive
            pty.eraseInLine(2);
            break;
         }
         break;
      // DA -- Device Attributes ESC [?1;0c
      case 'c':
         put(DA_RESPONSE);
         break;
      case 'g':
         switch (argn[0])
         {
         case 0:
            // clear stop at current position
            tabStops.remove(pty.getCursorCol());
            break;
         case 3:
            // clear all horizontal tab stops
            tabStops.clear();
            break;
         }
         break;
      // DECSTBM -- Set Top and Bottom Margins (DEC Private)
      case 'r':
         pty.setScrollRegion(argn[0], argn[1]);
         break;
      case 'n':
         switch (argn[0])
         {
         case 5:
            // DSR ready no malfunctions detected
            byte[] response = {ESC, '[', '0', 'n'};
            put(response);
            break;
         case 6:
            // CPR cursor position request
            byte[] data = String.format("E[%d;%dR", 
                                        pty.getCursorRow() + 1, 
                                        pty.getCursorCol() + 1).getBytes();
            data[0] = ESC;
            put(data);
            break;
         }
         break;
      // SGR -- Select Graphic Rendition
      case 'm':
         for (int i = 0; i <= argc; i++)
         {
            switch (argn[i])
            {
            case 0: // Reset all attributes
               // Attributes off
               attribute = pty.attributeMask(0);
               color = 0;
               break;
            case 1: // Bright // bold
               // Bold or increased intensity
               attribute |= pty.attributeMask(2);
               break;
            case 4: // Underline
               // Underscore
               attribute |= pty.attributeMask(3);
               break;
            case 5: // Blink slow
            case 6: // Blink rapid
               attribute |= pty.attributeMask(1);
               break;
            case 7: // reverse
               //enter_reverse_mode();
               attribute |= pty.attributeMask(4);
               break;
            case 30:
            case 31:
            case 32:
            case 33:
            case 34:
            case 35:
            case 36:
            case 37:
               // set FG color
               int fgColor = argn[i] - 30 + 1;
               color |= ((fgColor << 16) & 0xFFFF0000);
               break;
            case 40:
            case 41:
            case 42:
            case 43:
            case 44:
            case 45:
            case 46:
            case 47:
               // set BG color
               int bgColor = argn[i] - 40 + 1;
               color |= (bgColor & 0x0000FFFF);
               break;
            default:
               break;
            }
         }         
         break;
      case '?':
         int code = 0;
         // decode numeric parameters
         while (true)
         {
            b = this.nextByte();
            if (!isDigit(b))
            {
               break;
            }
            code = code * 10 + (b - 0x30);
         }
         switch (code)
         {
         case 1:
            // DECCKM
            if (b == 'l')
            {
               // Reset selects cursor keys to generate ANSI cursor control sequences.
               keyboard.setCursorKey(false);
            }
            else if (b == 'h')
            {
               // Set selects cursor keys to generate control (application) functions.
               keyboard.setCursorKey(true);
            }
            break;
         case 3:
            // DECCOLM
            if (b == 'l')
            {
               // 80 char
            }
            else if (b == 'h')
            {
               // 132 char
            }
            break;
         case 7:
            // DECAWM
            if (b == 'l')
            {
               // AutoWrap OFF
               pty.setAutoWrap(false);
            }
            else if (b == 'h')
            {
               // AutoWrap ON
               pty.setAutoWrap(true);
            }
            break;            
         case 25:
            // DECTCEM
            if (b == 'l')
            {
               // Cursor OFF
               pty.setCursorVisible(false);
            }
            else if (b == 'h')
            {
               // Cursor ON
               pty.setCursorVisible(true);
            }
            break;            
         }
         break;
      }
   }
   
   /**
    * Process VT52 mode
    */
   private void doVT52()
   {
      byte b = this.nextByte();
      switch (b)
      {
      case '<':   
         // SETANSI Enter/exit ANSI mode (VT52)
         mode = TerminalMode.ANSI;
         return;
      case 'A':
         // cursorup
         pty.moveCursorPos(-1, 0);
         break;
      case 'B':
         // cursordn
         pty.moveCursorPos(1, 0);
         break;
      case 'C':
         // cursorrt
         pty.moveCursorPos(0, 1);
         break;
      case 'D':
         pty.moveCursorPos(0, -1);
         // cursorlf
         break;
      case 'H':
         // cursorhome
         pty.setCursorPos(0, 0);
         break;
      case 'I':
         // revindex
         pty.reverseIndex();
         break;
      case 'J':
         // cleareos
         pty.eraseInDisplay(0);
         break;
      case 'K':
         // cleareol
         pty.eraseInLine(0);
         break;
      case 'Y':
         // cursorpos
         int row = this.nextByte() - 0x1f;
         int col = this.nextByte() - 0x1f;
         pty.setCursorPos(row, col);
         break;
      case 'Z':
         // ident
         this.put(IDENT_RESPONSE);
         break;
      default:
         break;
      }
   }

   @Override
   public void run()
   {
      this.running = true;
      this.doWork();
   }
   
   /**
    * Go to the next tab stop position.
    * If no more tab stops go to end of line.
    */
   private void goNextTabstop()
   {
      if (!tabStops.isEmpty())
      {
         int newPos = pty.getColumns() - 1;
         int curPos = pty.getCursorCol();
         // search next tab stop
         for (Integer tabStop : tabStops)
         {
            if (tabStop > curPos)
            {
               newPos = tabStop;
               break;
            }
         }
         pty.setCursorCol(newPos);
      }      
   }
   
   /**
    * Read an UTF-8 byte sequence transform to UTF-16 and store on a cell structure.  
    * 
    * @param   v0
    *          First value in stream;
    */
   private void putUtf8Stream(int v0)
   {
      ByteArrayOutputStream seq = new ByteArrayOutputStream();

      if (v0 <= 0x7f)
      {
         // Test if printable
         if (v0 > 0x1f)
         {
            seq.write(v0);    
         }
      }
      else
      {
         int v1 = this.nextByte() & 0xff;
         boolean c1 = (v1 <= 0xbf);
         if ((v0 <= 0xdf) && c1)
         {
            seq.write(v0);               
            seq.write(v1);          
         }
         else
         {
            int v2 = this.nextByte() & 0xff;
            boolean c2 = (v2 <= 0xbf);
            if ((v0 <= 0xef) && c1 && c2)
            {
               seq.write(v0);               
               seq.write(v1);               
               seq.write(v2);               
            } 
            else
            {
               int v3 = this.nextByte() & 0xff;
               boolean c3 = (v3 <= 0xbf);
               if ((v0 <= 0xf7) && c1 && c2 && c3)
               {
                  seq.write(v0);               
                  seq.write(v1);               
                  seq.write(v2);               
                  seq.write(v3);               
               }                      
            }
         }
      }
      
      // Set cell attributes
      pty.setCellText(convertUTF8toUTF16(seq), attribute, color);
   }
   
   /**
    * Convert a UTF-8 sequence to a UTF-16 character.
    * If sequence cannot be converted to UTF-16 replaced character is returned.
    *  
    * @param   utf8
    *          UTF-8 byte sequence.
    * 
    * @return  Converted UTF-16 character or 0 if cannot convert. 
    */
   private char convertUTF8toUTF16(ByteArrayOutputStream utf8)
   {
      /** UTF-16 replaced character */
      char utf16 = '\uFFFD';
      
      try
      {
         if (utf8.size() > 0)
         {
            String string = new String(utf8.toByteArray(), "UTF-8");
            utf16 = string.charAt(0);
         }
      }
      catch (UnsupportedEncodingException e)
      {
         e.printStackTrace();
      }
      
      return utf16;
   }

   /**
    * Test if byte is a ASCII decimal digit.
    * 
    * @param   b
    *          Input byte.
    *          
    * @return  <code>true</code> if is a digit false otherwise.
    */
   private boolean isDigit(byte b)
   {
      return (b >= (byte) 0x30) && (b <= (byte) 0x39);
   }
   
   /**
    * Request terminate. The thread should exit when no bytes are available
    * to read form master pty.
    */
   public void stopRunning()
   {
      terminate = true;
   }
}
