/*
** Module   : shell_linux.c
** Abstract : Run LINUX/UNIX shell and control master PTY descriptor.
**
** Copyright (c) 2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 MAG 20140318 Created initial version.
*/

#include <sys/types.h>
#include <sys/wait.h>
#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <pty.h>
#include <string.h>
#include <fcntl.h>

/**
 * Use non blocking operations on master PTY descriptor.
 * 
 * @param   mfd
 *          Master PTY file descriptor.
 *
 * @return  On error, -1 is returned, and errno is set appropriately.
 */
int unblock(int mfd)
{
   int flags;
   // Set non-blocking
   if ((flags = fcntl(mfd, F_GETFL, 0)) == -1)
   {
      flags = 0;
   }
   return fcntl(mfd, F_SETFL, flags | O_NONBLOCK); 
}

/**
 * Start a LINUX/UNIX shell session on a pseudo terminal. The master PTY file descriptor is used
 * to capture terminal INPUT and OUTOPUT streams.
 * 
 * @param   term
 *          A string to set terminal type environment.
 *          Example: "TERM=vt100"  
 * @param   cmdline
 *          Command line string.
 * @param   mfd_inp
 *          Pointer to a integer which store the master PTY descriptor.
 * @param   mfd_out
 *          Pointer to a integer which store the master PTY descriptor.
 * @param   overlay
 *          Overlapping mode. Not used on Linux always 1
 *
 * @return  On success return PID and set the master PTY descriptor.
 *          On error, -1 is returned, and errno is set appropriately.
 */
int os_shell(char* terminal, char** argv, int* mfd_inp, int *mfd_out, int overlay) 
{
   // PTY master descriptor.
   int mfd = -1;
   
   // Set TERM environment
   putenv(terminal);
   
   // Create child process
   int pid = forkpty(&mfd, NULL, NULL, NULL);   
   
   if (pid == 0) 
   {
      // child
      return execvp(*argv, argv);
   }

   if (pid > 0)
   {
      // Unblock
      unblock(mfd);      
   }
   
   // In Linux the INPUT and OUTPUT descriptors are the same
   *mfd_inp = mfd;
   *mfd_out = mfd;
   
   // parent
   return pid;
}

/**
 * Wait for a child process to exit.
 * 
 * @param   pid
 *          Child process pid.
 *
 * @return  Process status.
 *          On error, -1 is returned, and errno is set appropriately.
 */
int wait_pid(int pid)
{
   int status = -1;
   waitpid(pid, &status, 0);
   return status;
}

/**
 * Read from master PTY file descriptor.
 * 
 * @param   mfd
 *          Master PTY file descriptor.
 * @param   buffer
 *          The read buffer.
 * @param   size
 *          Maximum size to read.
 *
 * @return  Number of bytes read or -1 on error. 
 */
int read_pty(int mfd, char* buffer, int size)
{
   return read(mfd, buffer, size);
}

/**
 * Write to master PTY file.
 * 
 * @param   mfd
 *          Master PTY file descriptor.
 * @param   buffer
 *          Data to write.
 * @param   size
 *          Size of data to write.
 *
 * @return  Number of bytes written. 
 */
int write_pty(int mfd, void* buffer, int size)
{
   return write(mfd, buffer, (size_t) size);
}
