/*
** Module   : shell_win.c
** Abstract : Run WINDOWS shell "cmd.exe" and control STDIN, STDOUT and STDERR file descriptors.
**
** Copyright (c) 2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 MAG 20140318 Created initial version.
*/

void command_line(char* cmd, char** argv);
int os_shell_overlay(char* terminal, char** argv, int* mfd_inp, int* mfd_out);
int os_shell_no_overlay(char* terminal, char** argv, int* mfd_inp, int* mfd_out);

#include <windows.h>
#include <stdio.h>

#define MAX_COMMAND_LINE 8192

/**
 * Start a WINDOWS shell session. Simulate a Linux PTY using pipes. 
 * 
 * @param   term
 *          A string to set terminal type environment.
 *          Not used on Windows OS.
 * @param   cmdline
 *          Command line arguments.
 * @param   mfd_inp
 *          Pointer to a integer which store the STDIN handle.
 * @param   mfd_out
 *          Pointer to a integer which store the STDOUT, STDERR handles.
 * @param   overlay
 *          Overlapping mode.
 *
 * @return  On success return process handle and set standard files handles.
 *          On error, -1 is returned, and errno is set appropriately.
 */
int os_shell(char* terminal, char** argv, int* mfd_inp, int* mfd_out, int overlay)
{
   return overlay ? os_shell_overlay(terminal, argv, mfd_inp, mfd_out) :
                    os_shell_no_overlay(terminal, argv, mfd_inp, mfd_out);
} 

/**
 * Start a WINDOWS shell session. Simulate a Linux PTY using pipes. 
 * Use the same window as P2J client. 
 *
 * @param   term
 *          A string to set terminal type environment.
 *          Not used on Windows OS.
 * @param   cmdline
 *          Command line arguments.
 * @param   mfd_inp
 *          Pointer to a integer which store the STDIN handle.
 * @param   mfd_out
 *          Pointer to a integer which store the STDOUT, STDERR handles.
 *
 * @return  On success return process handle and set standard files handles.
 *          On error, -1 is returned, and errno is set appropriately.
 */
int os_shell_overlay(char* terminal, char** argv, int* mfd_inp, int *mfd_out) 
{
   // command line
   char cmd[MAX_COMMAND_LINE] = "";
   
   DWORD dwCreationFlags = CREATE_NEW_CONSOLE;
   HANDLE hReadInp = 0, hWriteInp = 0;
   HANDLE hReadOut = 0, hWriteOut = 0;
   HANDLE hWriteInpDup = 0;
   HANDLE hReadOutDup = 0;
      
   STARTUPINFO         si = {0};
   PROCESS_INFORMATION pi = {0};
   
   SECURITY_ATTRIBUTES sa = {0}; 
   sa.nLength = sizeof(SECURITY_ATTRIBUTES); 
   sa.bInheritHandle = TRUE; 
   sa.lpSecurityDescriptor = NULL; 

   // Indicate if pipes are created and configured
   int pipesOk = 0;

   // current process handle
   HANDLE hProcess = GetCurrentProcess();
   
   // Create pipe for STDIN
   if (CreatePipe(&hReadInp, &hWriteInp, &sa, 0))
   {
      if (DuplicateHandle(hProcess, 
                          hWriteInp, 
                          hProcess, 
                          &hWriteInpDup, 
                          0, 
                          FALSE, DUPLICATE_SAME_ACCESS))
      {
         // Create pipe for STDOUT and STDERR
         if (CreatePipe(&hReadOut, &hWriteOut, &sa, 0))
         {
            if (DuplicateHandle(hProcess, 
                                hReadOut, 
                                hProcess, 
                                &hReadOutDup, 
                                0, 
                                FALSE, DUPLICATE_SAME_ACCESS))
            {
               // pipes are created
               pipesOk = 1;

               // no longer nedded
               CloseHandle(hWriteInp);
               CloseHandle(hReadOut);
               
               // Prepare to start process
               si.cb = sizeof(STARTUPINFO);
               si.wShowWindow = SW_HIDE;
               si.dwFlags = STARTF_USESTDHANDLES | STARTF_USESHOWWINDOW;
               
               // Standard files redirection
               si.hStdInput = hReadInp;
               si.hStdOutput =  hWriteOut;
               si.hStdError = hWriteOut;
               
               // Build command line
               command_line(cmd, argv);
            }
         }
      }   
   }
      
   // Execute command
   if (pipesOk && CreateProcess(NULL,             // application name
                                cmd,              // command line
                                NULL,             // process security attributes
                                NULL,             // thread security attribute
                                TRUE,             // inherit handles
                                dwCreationFlags,  // flags
                                NULL,             // environment block
                                NULL,             // current directory
                                &si,              // startup info
                                &pi))             // process info
   {
      // Pipe handles
      *mfd_inp = HandleToLong(hReadOutDup);
      *mfd_out = HandleToLong(hWriteInpDup);
      // Close unused parent handles
      CloseHandle(pi.hThread);
      CloseHandle(hReadInp);
      CloseHandle(hWriteOut);
      // Return pid
      return HandleToLong(pi.hProcess);
   }
   else
   {
      // Clean up
      CloseHandle(hReadInp);
      CloseHandle(hWriteInp);
      CloseHandle(hWriteInpDup);
      CloseHandle(hReadOut);
      CloseHandle(hReadOutDup);
      CloseHandle(hWriteOut);
   }

   // On error return -1
   return -1;
}

/**
 * Start a WINDOWS shell session. Simulate a Linux PTY using pipes. 
 * Spawn on a separate window. 
 *
 * @param   term
 *          A string to set terminal type environment.
 *          Not used on Windows OS.
 * @param   cmdline
 *          Command line arguments.
 * @param   mfd_inp
 *          Pointer to a integer which store the STDIN handle.
 * @param   mfd_out
 *          Pointer to a integer which store the STDOUT, STDERR handles.
 *
 * @return  On success return process handle and set standard files handles.
 *          On error, -1 is returned, and errno is set appropriately.
 */
int os_shell_no_overlay(char* terminal, char** argv, int* mfd_inp, int* mfd_out) 
{
   // command line
   char cmd[MAX_COMMAND_LINE] = "";
   
   DWORD dwCreationFlags = CREATE_NEW_CONSOLE;
      
   STARTUPINFO         si = {0};
   PROCESS_INFORMATION pi = {0};
      
   si.cb = sizeof(STARTUPINFO);
   si.wShowWindow = SW_SHOW;
   si.dwFlags = STARTF_USESHOWWINDOW;
   
   // Build command line
   command_line(cmd, argv);
      
   // Execute command
   if (CreateProcess(NULL,             // application name
                     cmd,              // command line
                     NULL,             // process security attributes
                     NULL,             // thread security attribute
                     FALSE,            // inherit handles
                     dwCreationFlags,  // flags
                     NULL,             // environment block
                     NULL,             // current directory
                     &si,              // startup info
                     &pi))             // process info
   {
      // Close unused parent handles
      CloseHandle(pi.hThread);
      // Return pid
      return HandleToLong(pi.hProcess);
   }

   // On error return -1
   return -1;
}

/**
 * Wait for a child process to exit.
 * 
 * @param   pid
 *          Child process handle.
 *
 * @return  Process status.
 *          On error, a non zero value is returned, and errno is set appropriately.
 */
int wait_pid(int pid)
{
   HANDLE hProcess = LongToHandle(pid);
   // Wait for the process to finish with completion I/O operations
   WaitForSingleObjectEx(hProcess, INFINITE, TRUE); 
   return CloseHandle(hProcess);
}

/**
 * Read from master PTY file descriptor.
 * 
 * @param   mfd
 *          STDOUT, STDERR handle.
 * @param   buffer
 *          The read buffer.
 * @param   size
 *          Maximum size to read.
 *
 * @return  Number of bytes read or -1 on error. 
 */
int read_pty(int mfd, char* buffer, int size)
{
   DWORD dwAvail = 0, nNumberOfBytesRead = -1;
   HANDLE hPipe = LongToHandle(mfd);
   if (PeekNamedPipe(hPipe, NULL, 0, NULL, &dwAvail, NULL))
   {
      if (dwAvail)
      {   
         ReadFile(hPipe, buffer, size, &nNumberOfBytesRead, NULL);
      }
   }
   return nNumberOfBytesRead;
}

/**
 * Write to master PTY file.
 * 
 * @param   mfd
 *          STDINP handle.
 * @param   buffer
 *          Data to write.
 * @param   size
 *          Size of data to write.
 *
 * @return  Number of bytes written. 
 */
int write_pty(int mfd, void* buffer, int size)
{
   DWORD nNumberOfBytesWritten = -1;
   HANDLE hPipe = LongToHandle(mfd);
   WriteFile(hPipe, buffer, size, &nNumberOfBytesWritten, NULL);
   FlushFileBuffers(hPipe);
   return nNumberOfBytesWritten;
}

/**
 * Build command line.
 * 
 * @param   cmd
 *          Command line string.
 * @param   argv
 *          Command line arguments.
 */
void command_line(char* cmd, char** argv)
{
   // Build command line
   if (argv)
   {   
      int len = 0;
      char* arg = *argv;
      while (arg)
      {
         if (cmd[0])
         {
            strcat(cmd, " ");
            len++;
         }
         len += strlen(arg);
         if (len >= MAX_COMMAND_LINE)
            break;
         strcat(cmd, arg);
         arg = *(++argv);
      }
   }
}